/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.analysis.plugin.reader.mq;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.CountDownLatch;

import com.rabbitmq.client.Channel;
import com.rabbitmq.client.Connection;
import com.rabbitmq.client.ConnectionFactory;
import com.rabbitmq.client.ConsumerCancelledException;
import com.rabbitmq.client.QueueingConsumer;
import com.rabbitmq.client.ShutdownSignalException;

import kieker.analysis.IProjectContext;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.analysis.plugin.annotation.Property;
import kieker.analysis.plugin.reader.AbstractReaderPlugin;
import kieker.common.configuration.Configuration;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.common.record.IMonitoringRecord;
import kieker.analysis.plugin.reader.mq.Bits;

/**
 * Reads monitoring records from the queue of an established RabbitMQ connection.
 * 
 * @author Santje Finke
 * 
 * @since 1.8
 */
@Plugin(description = "A reader which reads records from a RabbitMQ queue",
		dependencies = "This plugin needs the file 'rabbitmq.client-*.jar'.",
		outputPorts = {
			@OutputPort(name = RabbitMQReader.OUTPUT_PORT_NAME_RECORDS, eventTypes = { IMonitoringRecord.class }, description = "Output Port of the JMSReader")
		},
		configuration = {
			@Property(name = RabbitMQReader.CONFIG_PROPERTY_NAME_PROVIDERURL, defaultValue = "localhost"),
			@Property(name = RabbitMQReader.CONFIG_PROPERTY_NAME_QUEUE, defaultValue = "analysis"),
			@Property(name = RabbitMQReader.CONFIG_PROPERTY_PASSWORD, defaultValue = "guest"),
			@Property(name = RabbitMQReader.CONFIG_PROPERTY_USER, defaultValue = "guest"),
			@Property(name = RabbitMQReader.CONFIG_PROPERTY_PORT, defaultValue = "5672")

		})
public final class RabbitMQReader extends AbstractReaderPlugin {

	/** The name of the output port delivering the received records. */
	public static final String OUTPUT_PORT_NAME_RECORDS = "monitoringRecords";
	/** The name of the configuration determining the RabbitMQ provider URL. */
	public static final String CONFIG_PROPERTY_NAME_PROVIDERURL = "mqProviderUrl";
	/** The name of the configuration determining the RabbitMQ Queue (e.g. queue1). */
	public static final String CONFIG_PROPERTY_NAME_QUEUE = "mqDestination";
	/** The username that is used to connect to a queue. */
	public static final String CONFIG_PROPERTY_USER = "guest";
	/** The password that is used to connect to a queue. */
	public static final String CONFIG_PROPERTY_PASSWORD = "guest";
	/** The port that is used to connect to a queue. */
	public static final String CONFIG_PROPERTY_PORT = "5672";

	static final Log LOG = LogFactory.getLog(RabbitMQReader.class); // NOPMD package for inner class

	private final String providerUrl;
	private final String queueName;
	private final String password;
	private final String username;
	private final int port;
	private Connection connection;
	private Channel channel;
	private ConnectionFactory factory;
	private QueueingConsumer normalConsumer;
	private QueueingConsumer registryConsumer;
	private final CountDownLatch cdLatch = new CountDownLatch(1);
	
	private final Map<Integer, String> stringRegistry = new HashMap<Integer, String>();

	/**
	 * Creates a new instance of this class using the given parameters.
	 * 
	 * @param configuration
	 *            The configuration used to initialize the whole reader. Keep in mind that the configuration should contain the following properties:
	 *            <ul>
	 *            <li>The property {@link #CONFIG_PROPERTY_NAME_PROVIDERURL}, e.g. {@code localhost}
	 *            <li>The property {@link #CONFIG_PROPERTY_NAME_QUEUE}, e.g. {@code queue1}
	 *            <li>The property {@link #CONFIG_PROPERTY_PASSWORD}, e.g. {@code password}
	 *            <li>The property {@link #CONFIG_PROPERTY_USER}, e.g. {@code username}
	 *            <li>The property {@link #CONFIG_PROPERTY_PORT}, e.g. {@code port}
	 *            </ul>
	 * @param projectContext
	 *            The project context for this component.
	 * 
	 * @throws IllegalArgumentException
	 *             If one of the properties is empty.
	 */
	public RabbitMQReader(final Configuration configuration, final IProjectContext projectContext) throws IllegalArgumentException {
		super(configuration, projectContext);

		// Initialize the reader bases on the given configuration.
		this.providerUrl = configuration.getStringProperty(CONFIG_PROPERTY_NAME_PROVIDERURL);
		this.queueName = configuration.getStringProperty(CONFIG_PROPERTY_NAME_QUEUE);
		this.username = configuration.getStringProperty(CONFIG_PROPERTY_USER);
		this.password = configuration.getStringProperty(CONFIG_PROPERTY_PASSWORD);
		this.port = configuration.getIntProperty(CONFIG_PROPERTY_PORT);
		// simple sanity check
		if ((this.providerUrl.length() == 0) || (this.queueName.length() == 0)) {
			throw new IllegalArgumentException("RabbitMQReader has not sufficient parameters. providerUrl ('" + this.providerUrl + "') or destination ('"
					+ this.queueName + "'), is null");
		}
		
		RabbitMQRegistryConsumer registryConsumer = new RabbitMQRegistryConsumer();
		registryConsumer.start();
	}

	/**
	 * Creates a new instance of this class using the given parameters.
	 * 
	 * @param configuration
	 *            The configuration used to initialize the whole reader. Keep in mind that the configuration should contain the following properties:
	 *            <ul>
	 *            <li>The property {@link #CONFIG_PROPERTY_NAME_PROVIDERURL}, e.g. {@code localhost}
	 *            <li>The property {@link #CONFIG_PROPERTY_NAME_QUEUE}, e.g. {@code queue1}
	 *            <li>The property {@link #CONFIG_PROPERTY_PASSWORD}, e.g. {@code password}
	 *            <li>The property {@link #CONFIG_PROPERTY_USER}, e.g. {@code username}
	 *            </ul>
	 * 
	 * @throws IllegalArgumentException
	 *             If one of the properties is empty.
	 * 
	 * @deprecated To be removed in Kieker 1.8.
	 */
	@Deprecated
	public RabbitMQReader(final Configuration configuration) throws IllegalArgumentException {
		this(configuration, null);
	}

	/**
	 * A call to this method is a blocking call.
	 * 
	 * @return true if the method succeeds, false otherwise.
	 */
	public boolean read() {
		boolean retVal = true;
		try {
			this.factory = new ConnectionFactory();
			this.factory.setHost(this.providerUrl);
			this.factory.setPort(this.port);
			this.factory.setConnectionTimeout(0);
			this.factory.setUsername(this.username);
			this.factory.setPassword(this.password);
			
			this.connection = this.factory.newConnection();
			this.channel = this.connection.createChannel();

			this.channel.queueDeclare(this.queueName, false, false, false, null);
			this.channel.queueDeclare("registryRecords", false, false, false, null);

			LOG.info("Listening to destination:" + this.queueName + " at " + this.providerUrl + " !\n***\n\n");

			this.normalConsumer = new QueueingConsumer(this.channel);
			this.channel.basicConsume(this.queueName, true, this.normalConsumer);
			
			this.registryConsumer = new QueueingConsumer(this.channel);
			this.channel.basicConsume("registryRecords", true, this.registryConsumer);

			while (!Thread.interrupted()) {

				if (!this.connection.isOpen() || !this.channel.isOpen()) {
					this.reconnect();
				}

				final QueueingConsumer.Delivery delivery = this.normalConsumer.nextDelivery();

				final ByteArrayInputStream bain = new ByteArrayInputStream(delivery.getBody());
				final ObjectInputStream in = new ObjectInputStream(bain);
				final Object message = in.readObject();

				if ((message instanceof IMonitoringRecord) && (!this.deliverIndirect(OUTPUT_PORT_NAME_RECORDS, message))) {
					LOG.error("deliverRecord returned false");
				}

			}

		} catch (final IOException ex) { // NOPMD NOCS (IllegalCatchCheck)
			LOG.error("Error in read()", ex);
			retVal = false;
		} catch (final ClassNotFoundException e) {
			LOG.error("Error in read(): ClassNotFound Exception", e);
			retVal = false;
		} catch (final ShutdownSignalException e) {
			LOG.error("Error in read(): ShutdownSignal Exception", e);
			retVal = false;
		} catch (final ConsumerCancelledException e) {
			LOG.error("Error in read(): ConsumerCancelled Exception", e);
			retVal = false;
		} catch (final InterruptedException e) {
			LOG.error("Error in read(): Interrupted Exception", e);
			retVal = false;
		} finally {
			try {
				if (this.connection != null) {
					this.connection.close();
				}
			} catch (final IOException e) {
				LOG.error("Error in read()", e);
			}
		}
		return retVal;
	}
	
	private IMonitoringRecord fromByteArray(final byte[] b) {
		final Class<?>[] recordTypes = monitoringRecord.getValueTypes();

		int arraySize = 4 + 8;

		for (int i = 0; i < recordTypes.length; i++) {
			if (recordTypes[i] == String.class) {
				arraySize += 4;
			} else if (recordTypes[i] == Integer.class
					|| recordTypes[i] == int.class) {
				arraySize += 4;
			} else if (recordTypes[i] == Long.class
					|| recordTypes[i] == long.class) {
				arraySize += 8;
			} else if (recordTypes[i] == Float.class
					|| recordTypes[i] == float.class) {
				arraySize += 4;
			} else if (recordTypes[i] == Double.class
					|| recordTypes[i] == double.class) {
				arraySize += 8;
			} else if (recordTypes[i] == Byte.class
					|| recordTypes[i] == byte.class) {
				arraySize += 1;
			} else if (recordTypes[i] == Short.class
					|| recordTypes[i] == short.class) {
				arraySize += 2;
			} else if (recordTypes[i] == Boolean.class
					|| recordTypes[i] == boolean.class) {
				arraySize += 1;
			} else {
				arraySize += 1;
			}
		}

		byte[] result = new byte[arraySize];
		int offset = 0;

		Bits.putInt(result, offset, this.monitoringController
				.getIdForString(monitoringRecord.getClass().getName()));
		offset += 4;

		Bits.putLong(result, offset, monitoringRecord.getLoggingTimestamp());
		offset += 8;
		final Object[] recordFields = monitoringRecord.toArray();

		for (int i = 0; i < recordFields.length; i++) {
			if (recordFields[i] == null) {
				if (recordTypes[i] == String.class) {
					Bits.putInt(result, offset,
							this.monitoringController.getIdForString(""));
					offset += 4;
				} else if (recordTypes[i] == Integer.class
						|| recordTypes[i] == int.class) {
					Bits.putInt(result, offset, 0);
					offset += 4;
				} else if (recordTypes[i] == Long.class
						|| recordTypes[i] == long.class) {
					Bits.putLong(result, offset, 0L);
					offset += 8;
				} else if (recordTypes[i] == Float.class
						|| recordTypes[i] == float.class) {
					Bits.putFloat(result, offset, 0);
					offset += 4;
				} else if (recordTypes[i] == Double.class
						|| recordTypes[i] == double.class) {
					Bits.putDouble(result, offset, 0);
					offset += 8;
				} else if (recordTypes[i] == Byte.class
						|| recordTypes[i] == byte.class) {
					Bits.putByte(result, offset, (byte) 0);
					offset += 1;
				} else if (recordTypes[i] == Short.class
						|| recordTypes[i] == short.class) {
					Bits.putShort(result, offset, (short) 0);
					offset += 2;
				} else if (recordTypes[i] == Boolean.class
						|| recordTypes[i] == boolean.class) {
					Bits.putBoolean(result, offset, false);
					offset += 1;
				} else {
					LOG.warn("Record with unsupported recordField of type "
							+ recordFields[i].getClass());
					Bits.putByte(result, offset, (byte) 0);
					offset += 1;
				}
			} else if (recordFields[i] instanceof String) {
				Bits.putInt(result, offset, this.monitoringController
						.getIdForString((String) recordFields[i]));
				offset += 4;
			} else if (recordFields[i] instanceof Integer) {
				Bits.putInt(result, offset, (Integer) recordFields[i]);
				offset += 4;
			} else if (recordFields[i] instanceof Long) {
				Bits.putLong(result, offset, (Long) recordFields[i]);
				offset += 8;
			} else if (recordFields[i] instanceof Float) {
				Bits.putFloat(result, offset, (Float) recordFields[i]);
				offset += 4;
			} else if (recordFields[i] instanceof Double) {
				Bits.putDouble(result, offset, (Double) recordFields[i]);
				offset += 8;
			} else if (recordFields[i] instanceof Byte) {
				Bits.putByte(result, offset, (Byte) recordFields[i]);
				offset += 1;
			} else if (recordFields[i] instanceof Short) {
				Bits.putShort(result, offset, (Short) recordFields[i]);
				offset += 2;
			} else if (recordFields[i] instanceof Boolean) {
				Bits.putBoolean(result, offset, (Boolean) recordFields[i]);
				offset += 1;
			} else {
				LOG.warn("Record with unsupported recordField of type "
						+ recordFields[i].getClass());
				Bits.putByte(result, offset, (byte) 0);
				offset += 1;
			}
		}

		return result;
	}
	
	/**
	 * Establishes a connection to a rabbitMQ channel with the current connection informationen.
	 */
	private void reconnect() {
		try {
			this.connection = this.factory.newConnection();
			this.channel = this.connection.createChannel();
			this.channel.queueDeclare(this.queueName, false, false, false, null);
			this.normalConsumer = new QueueingConsumer(this.channel);
			this.channel.basicConsume(this.queueName, true, this.normalConsumer);
		} catch (final IOException e) {
			RabbitMQReader.LOG.error("Error reestablishing connection", e);
		}
	}

	final void unblock() { // NOPMD (package visible for inner class)
		this.cdLatch.countDown();
	}

	final boolean deliverIndirect(final String outputPortName, final Object data) { // NOPMD (package visible for inner class)
		return super.deliver(outputPortName, data);
	}

	/**
	 * {@inheritDoc}
	 */
	public void terminate(final boolean error) {
		LOG.info("Shutdown of RabbitMQReader requested.");
		this.unblock();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Configuration getCurrentConfiguration() {
		final Configuration configuration = new Configuration();

		configuration.setProperty(CONFIG_PROPERTY_NAME_PROVIDERURL, this.providerUrl);
		configuration.setProperty(CONFIG_PROPERTY_NAME_QUEUE, this.queueName);
		configuration.setProperty(CONFIG_PROPERTY_PASSWORD, this.password);
		configuration.setProperty(CONFIG_PROPERTY_USER, this.username);

		return configuration;
	}
}
